/*
 * File:  main.cc
 * Author:  mikolas
 * Created on:  Mon, Jul 13, 2015 1:58:05 PM
 * Copyright (C) 2015, Mikolas Janota
 */
#include <signal.h>
#include "ReadQ.h"
#include "Options.h"
#include "QSolver.h"
#include "rareqs.h"
#include "auxiliary.h"
#include <stdlib.h>
#include "satsyn.h"
#include "defs.h"
using namespace std;
void read_qcir(string filename);
static void SIG_handler(int signum);
ostream& print_usage(const Options& options,ostream& o);
void log_header(const string& filename, const Prefix& pref, int argc, char const * const * argv);
void prn_args(ostream& o, int argc, char** argv);
int run_solver(QSolver* ps);
int run_solver(const Options& options, Rareqs* ps);
ostream& print_winning_move(ostream& o, const Move& wm);
void block_move(const Move& wm, Rareqs* ps);
QSolver* qps=NULL;
Rareqs* ps=NULL;

AigFactory factory;
extern int qcir_line;
extern QAigFla qcir_qfla;
extern unordered_map<const char*,int,cstrHash,cstrEq> name2var;
extern unordered_map<int,const char*> var2name;

void normalize_prefix(Prefix& p) {
    if (p.empty()) {
        p.push_back(Quantification(EXISTENTIAL, VarVector()));
        return;
    }
    size_t last = 0;// maitain the first quantifier
    for (size_t i = 1; i < p.size(); ++i) {
        auto& vs_i = p[i].second;
        if (vs_i.empty()) continue;
        if (p[i].first == p[last].first) {
            auto& vs_last = p[last].second;
            vs_last.insert(vs_last.end(), vs_i.begin(), vs_i.end());
        } else {
            ++last;
            if (last != i) p[last] = p[i];
        }
    }
    size_t new_sz = last+1;
    p.resize(new_sz);
}

int run_qcir(const string& flafile, Options& options, int argc, char** argv) {
   read_qcir(flafile);
   normalize_prefix(qcir_qfla.pref);
   //factory.print(cerr<<"read:\n", qcir_qfla.matrix)<<endl;
   //ps=new QSolver(qcir_qfla,options,factory);
   const bool rqs = qcir_qfla.pref.size() > 5 && var_count(qcir_qfla.pref) < 300;
   if (rqs) qps = new QSolver(qcir_qfla,options,factory);
   else ps = Rareqs::make_solver(options, factory, qcir_qfla.pref, qcir_qfla.matrix);
   LOG_LRN(log_header(flafile, qcir_qfla.pref, argc, argv););
   return rqs ? run_solver(qps) : run_solver(options, ps);
}

int run_cnf(const string& flafile, Options& options, int argc, char** argv) {
    scoped_ptr<Reader> fr;
    gzFile ff=Z_NULL;
    if (flafile.size()==1 && flafile[0]=='-') {
        fr.attach(new Reader(cin));
    } else {
        ff = gzopen(flafile.c_str(), "rb");
        if (ff == Z_NULL) {
            cerr << "ERROR: " << "Unable to open file: " << flafile << endl;
            cerr << "ABORTING" << endl;
            exit(100);
        }
        fr.attach(new Reader(ff));
    }
    ReadQ rq(*fr,false);
    try {
        rq.read();
    } catch (ReadException& rex) {
        cerr << "ERROR: " << rex.what() << endl;
        cerr << "ABORTING" << endl;
        exit(100);
    }
    cout<<"c done reading: "<<read_cpu_time()<<std::endl;
    if (!rq.get_header_read()) {
        cerr << "ERROR: Missing header." << endl;
        cerr << "ABORTING" << endl;
        exit(100);
    }
    LOG_LRN(log_header(flafile, rq.get_prefix(), argc, argv););

    QAigFla qf;
    AigUtil au(factory);
    qf.pref=rq.get_prefix();
    qf.matrix=au.convert(rq.get_clauses());
    if(qf.pref.size()==0) {
        qf.pref.push_back(make_pair(EXISTENTIAL,VarVector()));
    }
    //ps=new QSolver(qf, options, factory);
    ps=Rareqs::make_solver(options, factory, qf.pref, qf.matrix);
    return run_solver(options, ps);
}

int main(int argc, char** argv) {
  //test_satsyn(); return 0;
  srand(0);
  signal(SIGTERM, SIG_handler);
  signal(SIGINT, SIG_handler);
  signal(SIGABRT, SIG_handler);
#ifndef __MINGW32__
  signal(SIGHUP, SIG_handler);
  signal(SIGUSR1, SIG_handler);
#else
  cout << "c MINGW version." << endl;
#endif
#ifndef NDEBUG
  cout << "c DEBUG version." << endl;
#endif
  cout<<"c qfun, v00.0, "<<GITHEAD<<endl;
  cout<<"c (C) 2017 Mikolas Janota, mikolas.janota@gmail.com"<<endl;
#ifndef EXPERT
  // prepare nonexpert options
  const int nargc = 7;
  char* nargv[nargc];
  nargv[0] = argv[0];
  nargv[1] = strdup("-caps");
  nargv[2] = strdup("-i64");
  nargv[3] = strdup("-n4");
  nargv[4] = strdup("-b4");
  nargv[5] = strdup("-S4");
  nargv[6] = argc>=2 ? argv[1] : strdup("-");
  if (argc>2) {
    cerr<<"ERROR: ingoring some options after FILENAME"<<std::endl;
    return 100;
  }
  argv=nargv;
  argc=nargc;
#else
  cout<<"c WARNING: running in the EXPERT mode, I'm very stupid without any options."<<std::endl;
#endif
  //prn_args(cout << "c CMDS", argc, argv);
  Options options;
  if (!options.parse(argc, argv)) {
    cerr << "ERROR: processing options." << endl;
    print_usage(options,cerr);
    return 100;
  }
  auto& rest = options.get_rest();
  if (rest.size()>1) {
    cerr<<"ERROR: garbage at the end of command line."<<endl;
    return 100;
  }
#ifdef EXPERT
  //options.print_vals(cout);
#endif

  if (options.get_help()){
    print_usage(options,cout);
    return 0;
  }
  const string flafile(rest.size() ? rest[0] : "-");
  if (flafile=="-") cout<<"c reading from standard input"<<endl;
  else cout<<"c reading from "<<flafile<<endl;
  if (ends_with(flafile, ".qcnf") || ends_with(flafile, ".qdimacs")) {
      return run_cnf(flafile, options, argc, argv);
  } else {
      return run_qcir(flafile, options, argc, argv);
  }
}

static void SIG_handler(int signum) {
  cout<<"c received external signal " << signum << endl;
  if(ps){
    cout<<"c top level refine counter: "<<ps->get_refine_counter()<<std::endl;
    cout<<"c sum refine counter: "<<ps->get_sum_refine_counter()<<std::endl;
  }
  if (qps) {
      cout<<"c conflicts: "<<qps->get_conflicts()<<endl;
      qps->prn_refs(cout);
  }
  cout<<"c Terminating ..." << endl;
  LOG_LRN(lout<<"TIME: "<<read_cpu_time()<<endl;);
  LOG_LRN(lout<<"TIME/MEM OUT"<<endl;);
  exit(0);
}

ostream& print_usage(const Options& options,ostream& o) {
  o << "USAGE: [OPTIONS] [FILENAME]" << endl;
  return options.print(o);
}

void log_header(const string& filename, const Prefix& pref, int argc, char const * const *argv) {
    lout << "FILE:"<<filename<<endl;
    lout << "OPTIONS:";
    while(argc--) lout << " " << *(argv++);
    lout << endl;
    lout << "GITHEAD:"<<GITHEAD<<endl;
    lout << "PREFIX:"<<pref<<endl;
}

int run_solver(QSolver* ps) {
    const bool r=ps->solve();
    cout<<"c time: "<<read_cpu_time()<<std::endl;
    cout<<"c conflicts: "<<ps->get_conflicts()<<std::endl;
    cout<<"s cnf "<<(r?'1':'0')<<std::endl;
    LOG_LRN(lout<<"TIME: "<<read_cpu_time()<<endl;);
    LOG_LRN(lout<<"RESULT:"<<(r?"TRUE":"FALSE")<<endl;);
#ifdef DBG
    delete ps;
    ps = NULL;
#endif
    return r ? 10 : 20;
}

void log_computation(const Rareqs* ps) {
    cout<<"c TIME: "<<read_cpu_time()<<endl;
    cout<<"c top level refine counter: "<<ps->get_refine_counter()<<endl;
    cout<<"c sum refine counter: "<<ps->get_sum_refine_counter()<<endl;
    cout<<"c sum learning time: "<<ps->get_learning_time()<<endl;
}

int run_solver(const Options& options,Rareqs* ps) {
    const bool w=ps->wins();
    const bool r=(ps->quantifier_type()==EXISTENTIAL) == w;
    log_computation(ps);
    cout<<"s cnf "<<(r?'1':'0')<<std::endl;
    if (w) {
        Move wm;
        while (1) {
            wm.clear();
            ps->get_move(wm);
            print_winning_move(cout, wm);
            if (!options.get_win_mv_enum()) break;
            block_move(wm, ps);
            if (!ps->wins()) {
                cout<<"c enumeration completed"<<endl;
                break;
            }
            log_computation(ps);
        }
    }
    LOG_LRN(lout<<"RESULT:"<<(r?"TRUE":"FALSE")<<endl;);
#ifdef DBG
    delete ps;
    ps = NULL;
#endif
    return r ? 10 : 20;
}

void prn_args(ostream& o, int argc, char** argv) {
    for (int i = 0; i < argc; ++i) o << " " << argv[i];
    o << endl;
}

ostream& print_winning_move(ostream& o, const Move& wm) {
    o<<"v";
    for (Var v : ps->get_free()) {
        const lbool vv = eval(v, wm);
        if (vv == l_Undef) continue;
        const auto i = var2name.find(v);
        o<< " " << (vv==l_False ? '-' : '+');
        if (i == var2name.end()) o << v;
        else o << i->second;
    }
    return o<<endl;
}
void block_move(const Move& wm, Rareqs* ps) {
    LiteralVector blocking_clause;
    for (Var v : ps->get_free()) {
        const lbool vv = eval(v, wm);
        if (vv == l_Undef) continue;
        blocking_clause.push_back(vv == l_False ? mkLit(v) : ~mkLit(v));
    }
    LitSet blc = LitSet::mk(blocking_clause);
    //cerr<<"blocking_clause:"<<blc<<endl;
    ps->strengthen(blc);
}
